# coding: utf8

import argparse
import socket
import sys
import time
import xmlrpc.client
from pprint import pprint

import batou
import batou.component
import batou.environment
import batou.lib.file
import batou.template


class DNSAliases(batou.component.Component):
    """Set up DNS aliases under the .fcio.net domain.

    NOTE: Requires batou >= 1.5

    Environment configuration example::

        [host:your-host]
        data-alias-srv =
            your-alias
            your-other-alias
        data-alias-fe =
            portal

        # The DNSAliases component is needed only *once* per environmnent.
        components =
            dnsaliases

        [component:dnsaliases]
        # Optional postfix
        postfix = .foo

        # Project (or as it used to be called: resource group)
        project = demo

        # api key should go into secrets!
        api_key = your-api-key

    The aliases will be: `<alias-from-host><postfix><project>.fcio.net`. The
    example produces:

    * your-alias.foo.demo.fcio.net -- pointing to SRV
    * your-other-alias.foo.demo.fico.net -- pointing to SRV
    * portal.foo.demo.fcio.net -- pointing to FE

    The aliases will point to the SRV addresses.

    """

    _required_params_ = {"project": "demo"}
    postfix = ""
    project = None
    api_key = None

    # How long to wait for aliases (seconds). 0: do not wait
    wait_for_aliases = batou.component.Attribute(int, default=0)

    # class variable:
    calls = []

    def configure(self):
        if self.calls:
            return
        self._compute_calls()
        self._call()
        self._wait_for_aliases()

    def _compute_calls(self):
        assert self.project
        self.aliases = []
        for host in list(self.environment.hosts.values()):
            self._add_calls(host.name, "srv", host.data.get("alias-srv"))
            self._add_calls(host.name, "fe", host.data.get("alias-fe"))
        self.calls.sort(key=lambda c: c["name"])

    def _call(self):
        api = xmlrpc.client.ServerProxy(
            "https://{s.project}:{s.api_key}@api.flyingcircus.io/v1".format(
                s=self
            )
        )
        api.apply(self.calls)

    def _add_calls(self, hostname, interface, aliases_str):
        if not aliases_str:
            return
        aliases = aliases_str.split()
        aliases.sort()
        self.calls.append(
            {
                "__type__": "virtualmachine",
                "name": hostname + self.postfix,
                "aliases_" + interface: aliases,
            }
        )
        self.aliases.extend(aliases)

    def _wait_for_aliases(self):
        if not self.wait_for_aliases:
            return
        batou.output.line(
            "Waiting up to %s seconds for aliases." % self.wait_for_aliases
        )
        started = time.time()
        error = True
        while started + self.wait_for_aliases > time.time():
            error, results = self._check_aliases()
            for result in results:
                batou.output.line(result)
            if not error:
                break
            time.sleep(10)
        if error:
            raise RuntimeError("Aliases did not resolve in time.")

    def _check_aliases(self):
        error = False
        results = []
        for alias in self.aliases:
            fqdn = "{}{}.{}.fcio.net".format(alias, self.postfix, self.project)
            try:
                addrs = socket.getaddrinfo(fqdn, None, 0, 0, socket.IPPROTO_TCP)
            except socket.gaierror as e:
                result = str(e)
                error = True
            else:
                result = ", ".join(
                    sockaddr[0]
                    for (family, type, proto, canonname, sockaddr) in addrs
                )
            results.append("{}: {}".format(fqdn, result))
        return error, results


class Provision(batou.component.Component):
    """FCIO provisioning component.

    During a `batou deploy` run, this component does not do anything. It's
    there to hold the configuration.

    """

    project = None
    api_key = None
    location = "rzob"
    vm_environment_class = "NixOS"
    vm_environment = None
    api_url = "https://{project}:{api_key}@api.flyingcircus.io/v1"

    # Passed by the CLI runner, not meant to be set via environment:
    env_name: str = None
    diff: dict = None
    dry_run: bool = None

    def load_env(self):
        environment = batou.environment.Environment(self.env_name)
        environment.load()
        environment.load_secrets()
        if environment.exceptions:
            # Yeah, this is awkward.
            batou.output = batou._output.Output(batou._output.TerminalBackend())
            for exc in environment.exceptions:
                exc.report()
            sys.exit(1)
        return environment

    def get_api(self):
        rg_name = self.environment_.overrides["provision"]["project"]
        api_key = self.environment_.overrides["provision"]["api_key"]
        api_url = self.environment_.overrides["provision"].get("api_url")
        if not api_url:
            api_url = self.api_url
        api = xmlrpc.client.ServerProxy(
            api_url.format(project=rg_name, api_key=api_key)
        )
        return api

    def get_currently_provisioned_vms(self):
        return self.api.query("virtualmachine")

    def apply(self):
        self.environment_ = self.load_env()
        self.api = self.get_api()

        def config(name):
            value = self.environment_.overrides["provision"].get(name)
            if not value:
                value = getattr(self, name)
            return value

        rg_name = config("project")

        vms = []
        for name, host in sorted(self.environment_.hosts.items()):
            d = host.data
            roles = d.get("roles", "").splitlines()
            classes = ["role::" + r for r in roles if r]

            if d.get("environment", config("vm_environment")) is None:
                raise ValueError(
                    "'environment' for {} must be set.".format(name)
                )

            call = dict(
                __type__="virtualmachine",
                cores=int(d["cores"]),
                disk=max(int(d["disk"]), 30),
                memory=int(d["ram"]) * 1024,
                online=True,
                name=host.name,
                classes=classes,
                resource_group=rg_name,
                environment_class=d.get(
                    "environment_class", config("vm_environment_class")
                ),
                environment=d.get("environment", config("vm_environment")),
                location=config("location"),
                rbd_pool=d.get("rbdpool", "rbd.hdd"),
                frontend_ips_v4=int(d.get("frontend-ipv4", 0)),
                frontend_ips_v6=int(d.get("frontend-ipv6", 0)),
                service_description=d.get("description", ""),
            )

            def alias(interface):
                aliases = d.get("alias-" + interface)
                if aliases:
                    aliases = aliases.split()
                    aliases.sort()
                    call["aliases_" + interface] = aliases

            alias("srv")
            alias("fe")

            vms.append(call)

        if self.diff:
            diff = self.get_diff(self.get_currently_provisioned_vms(), vms)
            print(
                "Applying the configuration to {env} would yield the following"
                " changes:\n".format(env=self.env_name)
            )

            for vm, changes in sorted(diff.items()):
                if changes:
                    print(vm)
                    for key, (old, new) in sorted(changes.items()):
                        if old or new:
                            print(
                                "    {key:20}: {old} → {new}".format(**locals())
                            )
                        else:
                            print("    {key}".format(**locals()))
                else:
                    print("{vm}: <no changes>".format(vm=vm))
        else:
            serviceuser = dict(
                __type__="serviceuser",
                uid=self.environment_.service_user,
                resource_group=rg_name,
                description="Deployment service user",
            )
            calls = [serviceuser] + vms
            if self.dry_run:
                pprint(calls)
            else:
                pprint(self.api.apply(calls))

    def get_diff(self, old, new):
        result = {}

        old = {vm["name"]: vm for vm in old}
        new = {vm["name"]: vm for vm in new}

        for vm_name, old_vm in list(old.items()):
            result[vm_name] = changes = {}
            new_vm = new.get(vm_name)
            if not new_vm:
                changes["VM exists and is unknown to deployment"] = (None, None)
                continue
            # starting with new because that only includes the data we
            # can set. We ignore all the other keys.
            for key, new_value in list(new_vm.items()):
                old_value = old_vm.get(key)
                if key == "classes":
                    # Roles need special treatment, generic is always included
                    try:
                        old_value.remove("role::generic")
                    except ValueError:
                        pass
                if old_value != new_value:
                    changes[key] = (old_value, new_value)
        for vm_name, new_vm in list(new.items()):
            if vm_name in result:
                continue
            result[vm_name] = {"CREATE VM": (None, None)}
            result[vm_name].update(
                {key: (None, value) for key, value in list(new_vm.items())}
            )
        return result


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    p = subparsers.add_parser("provision", help="Apply resource settings")
    p.add_argument("env_name", help="Environment")
    p.add_argument("-n", "--dry-run", help="Dry run", action="store_true")
    p.add_argument(
        "-d", "--diff", help="Show changes in resources", action="store_true"
    )
    p.set_defaults(func=lambda **kw: Provision(**kw).apply())

    args = parser.parse_args()

    func_args = dict(args._get_kwargs())
    del func_args["func"]
    return args.func(**func_args)


if __name__ == "__main__":
    main()



from batou.component import Component, platform

from batou_ext.postfixadmin.dovecot import PFADovecot
from batou_ext.postfixadmin.postfix import PFAPostfix


@platform("gocept.net", PFADovecot)
class PFADovecotFC(Component):
    def verify(self):
        self.parent.assert_no_changes()

    def update(self):
        self.cmd("sudo -n /etc/init.d/dovecot reload")


@platform("gocept.net", PFAPostfix)
class PFAPostfixFC(Component):
    def verify(self):
        self.parent.assert_no_changes()

    def update(self):
        self.cmd("sudo -n /etc/init.d/postfix restart")



from textwrap import dedent

import batou.component
import batou.lib.file
import pkg_resources

import batou_ext.nix
import batou_ext.ssl


@batou_ext.nix.rebuild
class PostfixRelay(batou.component.Component):
    """Relay outpoing mail via another SMTP server.

    This requires the FC mailstub role and proper credentials for the remote
    host.

    """

    _required_params_ = {
        "smtp_relay_host": "smtp",
        "smtp_user": "scott",
        "smtp_password": "tiger",
    }
    smtp_relay_host = batou.component.Attribute(str)
    smtp_relay_port = batou.component.Attribute(int, 587)
    smtp_tls = batou.component.Attribute("literal", "True")

    smtp_auth = batou.component.Attribute("literal", default=True)
    smtp_user = batou.component.Attribute(str)
    smtp_password = batou.component.Attribute(str)

    provide_as = None  # (optional) str to self.provide()

    def configure(self):
        if self.provide_as:
            self.provide(self.provide_as, self)
        self.address = batou.utils.Address(self.host.fqdn, 25, require_v6=True)
        self += batou.lib.file.File(
            "/etc/local/postfix/main.cf",
            content=dedent(
                self.expand(
                    """
    relayhost = [{{component.smtp_relay_host}}]:{{component.smtp_relay_port}}
    {% if component.smtp_auth %}
    smtp_sasl_auth_enable = yes
    smtp_sasl_password_maps = hash:/etc/local/postfix/sasl_passwd
    smtp_sasl_security_options = noanonymous
    {% endif %}
    {% if component.smtp_tls %}smtp_use_tls = yes{% endif %}
    """
                )
            ),
        )

        if self.smtp_auth:
            self += batou.lib.file.File(
                "/etc/local/postfix/sasl_passwd",
                content=(
                    f"[{self.smtp_relay_host}]:{self.smtp_relay_port}"
                    f" {self.smtp_user}:{self.smtp_password}"
                ),
                sensitive_data=True,
            )
        else:
            self += batou.lib.file.Purge("/etc/local/postfix/sasl_passwd")
            self += batou.lib.file.Purge("/etc/local/postfix/sasl_passwd.db")

    def verify(self):
        if self.smtp_auth:
            self.assert_file_is_current(
                "/etc/local/postfix/sasl_passwd.db",
                requirements=["/etc/local/postfix/sasl_passwd"],
            )

    def update(self):
        if self.smtp_auth:
            with self.chdir("/etc/local/postfix"):
                self.cmd("postmap sasl_passwd")


@batou_ext.nix.rebuild
class Mailhog(batou.component.Component):
    """Set up a local testing mail server with mailog.

    This component provides a local dev mail setup.

    Usage:
    Just add mailhog to the enivronment and set `public_name`.
    Mailog will then be availabe under the `public_name` on port 80 and 443.
    You should use a subdomain to avoid conflicts, or change public_http(s).

    Note: Mails can be send to the FQDN of the host mailhog is running
    on. If you need: Connection details are stored inside the
    Address-object of self.address.

    For usage with basic auth you might like to use
    batou_ext.http.HTTPBasicAuth.

    When using the Mailhog component explicitly in one of your components you
    may do so like this::

        self.http_auth = self.require_one("http_basic_auth")
        self += batou_ext.mailhog.Mailhog(
            public_name=self.mail_public_name,
            http_auth_enable=True,
            http_basic_auth=self.http_auth
         )

    This example is adding an explicit batou_ext.http.HTTPBasicAuth
    rather than just pulling it from environment.
    """

    _required_params_ = {
        "public_name": "example.com",
        "public_smtp_name": "mail.flyingcircus.io",
    }
    public_name = batou.component.Attribute(str)
    public_smtp_name = batou.component.Attribute(str, default=None)
    mailport = batou.component.Attribute(int, 1025)
    uiport = batou.component.Attribute(int, 8025)
    apiport = batou.component.Attribute(int, 8025)
    purge_old_mailhog_configs = batou.component.Attribute(
        "literal", default=True
    )
    http_auth_enable = batou.component.Attribute("literal", default=False)
    http_basic_auth = None

    systemd_namespace = batou.component.Attribute(str, default="")
    disable_stdout = batou.component.Attribute("literal", default=False)

    # Either memory or maildir
    # mongodb is not yet supported
    storage_engine = batou.component.Attribute(str, default="memory")

    provide_as = None  # (optional) str to self.provide()

    def configure(self):
        if self.provide_as:
            self.provide(self.provide_as, self)

        if not self.public_smtp_name:
            self.public_smtp_name = self.host.fqdn

        # Migration from old nginx.conf
        if self.purge_old_mailhog_configs:
            self += batou.lib.file.Purge("mailhog")
            self += batou.lib.file.Purge("mailhog_env")
            self += batou.lib.file.Purge("/etc/local/systemd/mailhog.service")
            self += batou.lib.file.Purge("/etc/local/nginx/mailhog.conf")
            self += batou.lib.file.Purge("htdocs")

        self.address = batou.utils.Address(self.public_smtp_name, self.mailport)
        self.address_ui = batou.utils.Address(self.public_name, self.uiport)

        if self.http_auth_enable:
            if self.http_basic_auth is None:
                self.http_auth = self.require_one("http_basic_auth")
            else:
                self.http_auth = self.http_basic_auth

        self += batou.lib.file.File(
            "/etc/local/nixos/mailhog.nix",
            content=pkg_resources.resource_string(
                __name__, "resources/mailhog/mailhog.nix"
            ),
        )


@batou_ext.nix.rebuild
class Mailpit(batou.component.Component):
    _required_params_ = {
        "public_name": "example.com",
    }

    public_name = batou.component.Attribute(str)
    public_smtp_name = batou.component.Attribute(str, default=None)
    ui_port = batou.component.Attribute(int, 8025)
    smtp_port = batou.component.Attribute(int, 1025)

    max = batou.component.Attribute(int, 500)

    http_basic_auth = None

    provide_as = None  # (optional) str to self.provide()

    def configure(self):
        if self.provide_as:
            self.provide(self.provide_as, self)

        if not self.public_smtp_name:
            self.public_smtp_name = self.host.fqdn

        self.address = batou.utils.Address(
            self.public_smtp_name, self.smtp_port
        )

        if self.http_basic_auth is None:
            self.http_auth = self.require_one("http_basic_auth")
        else:
            self.http_auth = self.http_basic_auth

        self += batou.lib.file.File(
            "/etc/local/nixos/mailpit.nix",
            content=pkg_resources.resource_string(
                __name__, "resources/mailpit.nix"
            ),
        )



import os

import batou.component
import batou.lib.file
import batou.utils

import batou_ext.nix


@batou_ext.nix.rebuild
class Redis(batou.component.Component):
    """Component to define Redis password and address"""

    _required_params_ = {
        "password": "tiger",
    }
    password = None

    password_file = "/etc/local/redis/password"
    port = 6379

    # Number of database you are connecting
    db = batou.component.Attribute(int, default=0)

    cleanup = batou.component.Attribute("literal", default=False)
    cleanup_command = batou.component.Attribute(str, default="FLUSHDB")

    provide_as = batou.component.Attribute(str, default="redis")

    def configure(self):
        assert self.password
        self.provide(self.provide_as, self)

        self.address = batou.utils.Address(self.host.fqdn, self.port)
        self += batou.lib.file.File(
            self.password_file, content=self.password, sensitive_data=True
        )
        if self.cleanup:
            self += batou.lib.file.File(
                "rediscleanup.sh",
                source=self.resource("rediscleanup.sh"),
                mode=0o755,
            )
            self += batou_ext.run.Run("rediscleanup.sh", file=self._)

    def resource(self, filename):
        return os.path.join(os.path.dirname(__file__), "resources", filename)



import batou.component
import batou.lib.file
import pkg_resources


class JournalBeatTransport(batou.component.Component):

    """A shortcut for sending logmessages via journalbeat to a centralised loghost."""

    nix_file_path = batou.component.Attribute(
        str, default="/etc/local/nixos/journalbeat.nix"
    )
    transport_name = batou.component.Attribute(str)
    graylog_host = batou.component.Attribute(str)
    graylog_port = batou.component.Attribute(int, default=12301)

    def configure(self):

        self += batou.lib.file.File(
            self.nix_file_path,
            content=pkg_resources.resource_string(
                __name__, "resources/journalbeat.nix"
            ),
        )



"""Helper for Jenkins pipeline deployments."""

import argparse
import configparser
import json
import subprocess
import sys


def git_ls_remote(url, ref):
    cmd = subprocess.Popen(
        ["git", "ls-remote", url, ref], stdout=subprocess.PIPE
    )
    stdout, _ = cmd.communicate()
    if cmd.returncode != 0:
        raise ValueError(
            f"'git ls-remote {url} {ref}' failed with exit code {cmd.returncode}. Stderr is printed above."
        )
    if stdout:
        return stdout.decode("ascii").split("\t", 1)


def git_resolve(url, version):
    if len(version) == 40:
        # revision.
        try:
            int(version, 16)
        except ValueError:
            pass
        else:
            return version

    ls_remote = git_ls_remote(url, version)
    if ls_remote:
        rev, full_refspec = ls_remote
        # If full_refspec indicates that a tag was found, retry with `^{}`
        # to check if the tag is annotated: an annotated tag (e.g. due to its
        # own message or a signature) has its own revision which is not equal
        # to the revision of the commit that got tagged.
        # The difference is crucial in some cases however, e.g. if the commit's
        # rev is part of the S3 URL containing the frontend artifacts for this
        # deploy.
        if full_refspec.startswith("refs/tags/"):
            ls_remote_tag = git_ls_remote(url, f"{version}^{{}}")
            # If the result is empty here, the tag is not annotated, thus using the
            # rev from the first invocation is correct. We don't end up here because of
            # an error anymore since non-zero returns now fail with a ValueError earlier.
            if ls_remote_tag:
                rev = ls_remote_tag[0]
        return rev


class VersionsUpdater:
    UPDATERS = {
        "git-resolve": "update_git",
        "pass": "update_pass_value",
    }

    def __init__(self, versions_file, version_mapping_json):
        self.version_mapping = json.loads(version_mapping_json)
        self.versions_file = versions_file
        self.config = configparser.ConfigParser()
        self.config.read(self.versions_file)

    def __call__(self):
        for service, version in sorted(self.version_mapping.items()):
            if not version:
                # leave empty to keep current version
                continue
            self.update(service, version)

        with open(self.versions_file, "w") as f:
            self.config.write(f)
            # Remove the trailing newline, which pre-commit doesn't like:
            f.truncate(f.tell() - 1)

    def update(self, service, version):
        update_mode = self.config[service].get("update", "git-resolve")
        update_mode = update_mode.split(":", 1)
        mode = update_mode[0]
        args = "".join(update_mode[1:])

        func = getattr(self, self.UPDATERS[mode])
        func(service, version, args)

    def update_git(self, service, version, extra_args):
        resolved = git_resolve(self.config.get(service, "url"), version)
        if not resolved:
            raise ValueError(
                "%s: Could not resolve version %s." % (service, version)
            )
        log("%s: resolved version %s to: %s", service, version, resolved)
        self.config.set(service, "revision", resolved)
        self.config.set(service, "version", version)

    def update_pass_value(self, service, version, extra_args):
        self.config[service][extra_args] = version


def log(msg, *args):
    print(msg % args)
    sys.stdout.flush()


def list_components(versions_file, verbose=False):
    config = configparser.SafeConfigParser()
    config.read(versions_file)
    components = sorted(config.sections())
    if verbose:
        result = []
        for component in components:
            c = dict(config.items(component))
            c["name"] = component
            result.append(c)
    else:
        result = components

    print(json.dumps(result, sort_keys=True))


def set_versions(versions_file, version_mapping_json):
    vu = VersionsUpdater(versions_file, version_mapping_json)
    vu()


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    p = subparsers.add_parser(
        "list-components",
        help="List available components where versions can be set",
    )
    p.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Return all options from versions.ini, not only component names",
    )
    p.add_argument("versions_file", help='Name of "versions.ini"')
    p.set_defaults(func=list_components)

    p = subparsers.add_parser("set-versions", help="Update versions")
    p.add_argument(
        "versions_file",
        help="Name of versions.ini. If exists it will be overwritten.",
    )
    p.add_argument(
        "version_mapping_json", help="JSON: mapping of service: version"
    )
    p.set_defaults(func=set_versions)

    args = parser.parse_args()
    func_args = dict(args._get_kwargs())
    del func_args["func"]
    return args.func(**func_args)


if __name__ == "__main__":
    main()



import importlib
import pkgutil
from typing import List
from unittest.mock import Mock

from batou.component import Component

import batou_ext


def find_components_in(path, parents=()) -> List[object]:
    """Find classes deriving from `batou.component.Component` in given path.

    Return a list containing the classes.
    """
    classes = []
    modules = pkgutil.iter_modules(path)
    for _, name, ispkg in modules:
        if name == "tests":
            continue

        dotted_path = ["batou_ext"]
        dotted_path.extend(parents)
        dotted_path.append(name)

        module = importlib.import_module(".".join(dotted_path))
        if ispkg:
            classes.extend(
                find_components_in(module.__path__, parents=parents + (name,))
            )

        for name in dir(module):
            obj = getattr(module, name)
            if not isinstance(obj, type):
                continue
            if not issubclass(obj, Component):
                continue
            if obj.__module__.startswith("batou.lib."):
                # Omit from-imports of classes from batou.lib.*
                continue
            classes.append(obj)

    return classes


def pytest_generate_tests(metafunc):
    """Generate a test for each component."""
    classes = find_components_in(batou_ext.__path__)
    idlist = [dotted_name(x) for x in classes]
    argvalues = [(x,) for x in classes]

    metafunc.parametrize(
        ("component",), argvalues, ids=idlist, scope="function"
    )


def dotted_name(cls):
    return f"{cls.__module__}.{cls.__name__}"


class MockPostfixProvider(Component):
    """Mock component providing the expectations of postfix components."""

    address = Mock()
    connect = Mock()
    crt_file = Mock()
    database = Mock()
    dbms = Mock()
    key_file = Mock()
    password = Mock()
    username = Mock()

    def configure(self):
        self.provide("keypair::mail", self)
        self.provide("pfa::database", self)
        self.provide("postfix", self)
        self.provide("postgres", self)


class MockRoundcubeProvider(Component):
    """Mock component providing the expectations of roundcube components."""

    address = Mock()
    connect = Mock()
    database = Mock()
    dbms = Mock()
    password = Mock()
    username = Mock()

    def configure(self):
        self.provide("postfix", self)
        self.provide("postgres", self)
        self.provide("roundcube::database", self)


def test_prepare(root, mocker, component, tmpdir):
    """Assert that the `prepare` method of batou_ext components can be called.

    The `prepare` method itself calls the `configure` method.
    """
    mocker.patch("socket.gethostbyname", return_value="localhost")
    args = ("namevar",) if component.namevar else ()
    required = getattr(component, "_required_params_", None)
    kw = required if required else {}

    instance = component(*args, **kw)
    component_name = dotted_name(component)

    if component_name in {
        # expecting parent component to have a `crontab` attribute:
        "batou_ext.nix.InstallCrontab",
    }:
        root.crontab = "crontab"
    elif component_name in {
        # expecting a service user to be set and a `logrotate_conf`:
        "batou_ext.nix.LogrotateIntegration",
    }:
        root.environment.service_user = "service"
        root.logrotate_conf = Mock()
        root.logrotate_conf.content = "logrotate_conf content"
    elif component_name in {
        # expecting `executable`, `systemd`, and `checksum` attributes on
        # parent:
        "batou_ext.nix.UserInit",
    }:
        root.executable = "my_script"
        root.systemd = {}
        root.checksum = 42
    elif component_name in {"batou_ext.nix.UserEnv"}:
        instance.channel = "test-channel"
    elif component_name in {
        # expecting that some objects can be required:
        "batou_ext.postfixadmin.dovecot.PFADovecot",
        "batou_ext.postfixadmin.postfix.PFAPostfix",
        "batou_ext.postfixadmin.postgres.PFADatabase",
        "batou_ext.postfixadmin.PFA",
    }:
        MockPostfixProvider().prepare(root)
    elif component_name in {
        # expecting that some objects can be required:
        "batou_ext.roundcube.postgres.RoundcubeDatabase",
        "batou_ext.roundcube.Roundcube",
    }:
        MockRoundcubeProvider().prepare(root)
    elif component_name in {
        # expecting to be able to connect to Amazon S3 with actual
        # credentials:
        "batou_ext.s3.S3",
    }:
        mocker.patch("boto3.resource")
    elif component_name in {
        "batou_ext.fcio.DNSAliases",
    }:
        instance._call = lambda: None
    elif component_name in {
        "batou_ext.journalbeat.JournalBeatTransport",
    }:
        instance.transport_name = "my_transport_name"
        instance.graylog_host = "my_graylog_host"
        instance.graylog_port = "my_graylog_port"
    elif component_name == "batou_ext.nix.NixFile":
        instance.content = ""
    elif component_name == "batou_ext.nixos.NixOSModuleContext":
        instance.source_component = instance
        root.environment.deployment_base = "/does-not-exist"
    elif component_name == "batou_ext.nixos.NixOSModule":
        # This cannot be used as root component and the setup is all wrong.
        return
        # instance.name = "mymod"
        # (tmpdir / "mymod.nix").write_text("{}", encoding="US-ASCII")
    elif component_name == "batou_ext.http.HTTPServiceWatchdog":
        instance.script = "/srv/s-myuser/start-application.sh"
        instance.watchdog_script_path = "/srv/s-myuser/watchdog-wrapper.py"
        instance.healthcheck_url = "https://example.com"
    elif component_name == "batou_ext.mail.Mailpit":
        batou_ext.http.HTTPBasicAuth(
            **batou_ext.http.HTTPBasicAuth._required_params_
        ).prepare(root)
    instance.prepare(root)



import argparse
import configparser
import json
import os
import os.path
import subprocess
import sys

import batou.environment
from InquirerPy import inquirer
from prompt_toolkit.completion import FuzzyWordCompleter, ThreadedCompleter

from batou_ext.jenkins import VersionsUpdater


def get_git_version_completer(url):
    def get_words():
        cmd = subprocess.Popen(
            ["git", "ls-remote", url], stdout=subprocess.PIPE, encoding="UTF-8"
        )
        stdout, stderr = cmd.communicate()
        words = []
        for line in stdout.splitlines():
            words.extend(line.split())
        return words

    return ThreadedCompleter(FuzzyWordCompleter(get_words))


def find_versions_ini(environment: batou.environment.Environment):
    if "versions_ini" in environment.overrides["settings"]:
        return environment.overrides["settings"]["versions_ini"]
    elif os.path.exists("versions.ini"):
        return "versions.ini"
    else:
        raise ValueError(
            "No versions.ini specified in the environment file (via `settings.versions_ini`) and the file `versions.ini` does not exist in the git root, cannot proceed"
        )


def select_versions_interactive(basedir: str) -> dict:
    envs = os.listdir(os.path.join(basedir, "environments"))
    envs.sort()
    environment_name = inquirer.select(
        message="Select environment to update",
        choices=envs,
    ).execute()

    environment = batou.environment.Environment(environment_name)
    environment.load()

    versions = configparser.ConfigParser()
    versions_ini = find_versions_ini(environment)
    versions.read(os.path.join(basedir, versions_ini))

    components = versions.sections()
    selected_components = inquirer.select(
        message="Select components to update (Space to select, Ctrl-A for all):",
        choices=components,
        multiselect=True,
    ).execute()

    interative_versions = {}
    for c in selected_components:
        try:
            default = versions.get(c, "default")
        except configparser.NoOptionError:
            default = ""

        git_url = versions[c].get("url")
        if git_url:
            git_completer = get_git_version_completer(git_url)
        else:
            git_completer = None
        new_version = inquirer.text(
            message=f"Update {c} to:",
            default=default,
            completer=git_completer,
        ).execute()

        interative_versions[c] = new_version

    set_versions(versions_ini, interative_versions)
    return interative_versions


def get_current_versions(
    basedir: str, environment: batou.environment.Environment
):
    versions = configparser.ConfigParser()
    versions_ini = find_versions_ini(environment)
    versions.read(os.path.join(basedir, versions_ini))

    target_versions = {}
    for section in versions.sections():
        try:
            target_versions[section] = versions.get(section, "default")
        except configparser.NoOptionError:
            pass
    return target_versions


def update_from_branch(basedir: str, branch: str):
    branch_is_used = False
    envs = sorted(os.listdir(os.path.join(basedir, "environments")))

    for env_name in envs:
        environment = batou.environment.Environment(env_name)
        environment.load()

        if environment.branch != branch:
            continue

        branch_is_used = True

        print(f"Updating environment: {env_name}")
        current_versions = get_current_versions(basedir, environment)
        versions_ini = find_versions_ini(environment)
        set_versions(versions_ini, current_versions)

    if not branch_is_used:
        raise ValueError(
            f"Branch {branch} is not used in any environment, cannot auto-upate."
        )


def update_single(basedir: str, environment_name: str):
    environment = batou.environment.Environment(environment_name)
    environment.load()
    versions_ini = find_versions_ini(environment)
    current_versions = get_current_versions(basedir, environment)
    set_versions(versions_ini, current_versions)


def set_versions(versions_ini: str, target_versions: dict):
    updater = VersionsUpdater(versions_ini, json.dumps(target_versions))
    updater()
    subprocess.run(["git", "-P", "diff", versions_ini])


def main():
    basedir = os.environ.get("APPENV_BASEDIR", os.path.dirname(sys.argv[0]))

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a",
        "--update-defaults",
        dest="environment",
        help="Auto-update all versions to their specified defaults.",
    )
    parser.add_argument(
        "-b",
        "--branch",
        dest="branch",
        help="Update environment matching the branch.",
    )

    parser.add_argument(
        "-d",
        "--base-dir",
        dest="basedir",
        help="Deployment base dir",
        default=basedir,
    )

    args = parser.parse_args()

    if args.environment:
        update_single(basedir, args.environment)
    elif args.branch:
        update_from_branch(basedir, args.branch)
    else:
        select_versions_interactive(basedir)


if __name__ == "__main__":
    main()
